/**
 *	Polym sections
 *	Implements the protection of "core"
 **/

#include <windows.h>
#include <stdlib.h>
#include <time.h>

#include "polym.h"
#include "library.h"

#include "rc4.h"
#include "aPlib\\lib\\mscoff\\aplib.h"

#pragma warning(disable: 4312 4311)

#define MAKE_PTR(x,y) (LPVOID)((DWORD)x + y)

typedef struct section_mirror
	{
		IMAGE_SECTION_HEADER BaseHeader;
		IMAGE_SECTION_HEADER NewHeader;
	} SECTION_MIRROR, *PSECTION_MIRROR;

DWORD	key[] = {	0x0,	0x0,	0x0,	0x0	};

DWORD ConvertAddr(PSECTION_MIRROR mirror, WORD size, DWORD VirtualAddr)
{
	DWORD result = 0;

	for(WORD i = 0; i < size && !result; i++)
	{
		if (VirtualAddr >= mirror[i].BaseHeader.VirtualAddress && 
			VirtualAddr <= (mirror[i].BaseHeader.VirtualAddress + mirror[i].BaseHeader.Misc.VirtualSize))
		{
			result = (VirtualAddr - mirror[i].BaseHeader.VirtualAddress)
				+ mirror[i].NewHeader.VirtualAddress;
		}
	}

	return result;
}

/**
  *	VirtualAddr = RelocTable
  *	sizeReloc = Reloc Size
  * newBaseImage = new base image
  * oldBaseImage = old base image
  * lpBaseAddr = code
 **/
void ModifyReloc(LPVOID lpBaseAddr, DWORD oldBaseImage, DWORD newBaseImage, LPVOID VirtualAddr, DWORD sizeReloc)
{
#ifdef DEBUG
	OutputDebugStr("\"ModifyReloc\"");
#endif

	LPBYTE ptr = (LPBYTE) VirtualAddr;

#ifdef DEBUG
	char tmpOutput[128];
#endif
	while(sizeReloc > 0)
	{
		DWORD *entry = (DWORD *)(ptr);
		WORD *entries = (WORD *)(ptr);

		DWORD symbols = entry[1];
		DWORD baseRva = entry[0];
		DWORD newBaseRva = baseRva;

		#define REASSIGN(x) x = (x - oldBaseImage + newBaseImage)

		//REASSIGN(*entry);
		//REASSIGN(newBaseRva);

#ifdef DEBUG		
		wsprintf(tmpOutput, "-> Addr: %08x:%04x\t%04x\n", entry, symbols, sizeReloc);
		OutputDebugStr(tmpOutput);
#endif

		if (symbols > 8)
		{
			symbols -= 8;
			symbols /= 2;

			for(DWORD i = 0; i < symbols; i++)
			{
				WORD offset = entries[4+i] & 0xFFF;
				WORD type = (entries[4+i] & 0xF000) >> 12;
				DWORD *dwOpcode = NULL;
				DWORD addr = 0;
				DWORD newValue;

				switch(type)
				{
					case 3:
							addr = baseRva + offset;
							//REASSIGN(addr);

							dwOpcode = (DWORD *) MAKE_PTR(lpBaseAddr, addr);
							
							newValue = *dwOpcode; // old offset
							newValue -= oldBaseImage;
							newValue += newBaseImage;
							//newValue -= baseRva;
							//newValue += newBaseRva;

							*dwOpcode = newValue;

						break;
					default:
							OutputDebugStringA("Unmanaged reloc symbol");
						break;
				}

				// fix symbol!

				//entries[4+i]	// 4 is base!
			}
		}

		#undef REASSIGN
		ptr += entry[1];
		sizeReloc -= entry[1];
	}
}

void AdjustRELOC(LPVOID lpBaseAddr, DWORD oldBaseImage, DWORD newBaseImage, PSECTION_MIRROR mirror, WORD size, LPVOID VirtualAddr, DWORD sizeReloc)
{
	LPBYTE ptr = (LPBYTE) VirtualAddr;

	while(sizeReloc > 0)
	{
		DWORD *entry = (DWORD *)(ptr);

		WORD *entries = (WORD *)ptr;

		DWORD symbols = entry[1];
		DWORD baseRva = entry[0];	// base RVA
		DWORD newBaseRva = entry[0];

		#define REASSIGN(x) x = ConvertAddr(mirror, size, x)
		REASSIGN(*entry);	// reassign entry
		REASSIGN(newBaseRva);

		if (symbols > 8)
		{
			symbols -= 8;
			symbols /= 2;

			for(DWORD i = 0; i < symbols; i++)
			{
				WORD offset = entries[4+i] & 0xFFF;
				WORD type = (entries[4+i] & 0xF000) >> 12;
				DWORD *dwOpcode = NULL;
				DWORD addr = 0;
				DWORD newValue;

				switch(type)
				{
			
				case 3:
							addr = baseRva + offset;
							REASSIGN(addr);

							dwOpcode = (DWORD *) MAKE_PTR(lpBaseAddr, addr);
							
							newValue = *dwOpcode; // old offset
							newValue -= oldBaseImage;
							newValue += newBaseImage;
							newValue -= baseRva;
							newValue += newBaseRva;

							*dwOpcode = newValue;

						break;
					default:
							OutputDebugStringA("Unmanaged reloc symbol");
						break;
				}

				// fix symbol!

				//entries[4+i]	// 4 is base!
			}
		}

		#undef REASSIGN

		ptr += entry[1];
		sizeReloc -= entry[1];
	}

}

void AdjustIATEntry(PSECTION_MIRROR mirror, WORD size, LPVOID VirtualAddr)
{
	PDWORD entry = (PDWORD) VirtualAddr;

	while(*entry != 0)
	{
		if ((*entry & 0x80000000) == 0x00)
		{
#define REASSIGN(x) x = ConvertAddr(mirror, size, x)
			REASSIGN(*entry);
#undef REASSIGN
		}

		entry++;
	}

}

void AdjustIAT(LPVOID lpBaseAddr, PSECTION_MIRROR mirror, WORD size, LPVOID VirtualAddr, DWORD descriptor_size)
{
	for(WORD i = 0; i < (descriptor_size / sizeof(IMAGE_IMPORT_DESCRIPTOR)); i++)
	{
		PIMAGE_IMPORT_DESCRIPTOR descriptor = (PIMAGE_IMPORT_DESCRIPTOR)VirtualAddr;

		if (descriptor[i].Characteristics == 0)
			continue;

#define REASSIGN(x) x = ConvertAddr(mirror, size, x)

		REASSIGN(descriptor[i].Name);
		REASSIGN(descriptor[i].FirstThunk);
		REASSIGN(descriptor[i].OriginalFirstThunk);
#undef REASSIGN

		AdjustIATEntry(mirror, size, MAKE_PTR(lpBaseAddr, descriptor[i].OriginalFirstThunk));
	}
}

extern VOID ModifyExport( LPVOID lpBase, DWORD address, DWORD len);

void _fix_relocation_table()
{
	
}

/**
 *	_clone_section
 *	Create a duplicate of a section from an image PE to another IMAGE pe
 *	adding the section in headers with a new base address
 **/
PIMAGE_SECTION_HEADER _clone_section(
		LPVOID lpDestAddr, 
		PIMAGE_SECTION_HEADER lpLastDestSection, 
		PIMAGE_NT_HEADERS lpHeader, 
		WORD count,
		PIMAGE_SECTION_HEADER lpSrcSection,
		LPVOID lpSrcAddr)
{
	PIMAGE_SECTION_HEADER pHermit = AddSection(lpDestAddr);

	strcpy_s((char*) pHermit->Name, IMAGE_SIZEOF_SHORT_NAME, ".hermit");
	
	pHermit->Name[7] = '0' + count;

	pHermit->Characteristics = lpSrcSection->Characteristics;
	pHermit->NumberOfLinenumbers = lpSrcSection->NumberOfLinenumbers;
	pHermit->NumberOfRelocations = lpSrcSection->NumberOfRelocations;
	pHermit->PointerToLinenumbers = lpSrcSection->PointerToLinenumbers;
	pHermit->PointerToRawData = 
		RoundUp(lpLastDestSection->PointerToRawData + lpLastDestSection->SizeOfRawData, lpHeader->OptionalHeader.FileAlignment);

	pHermit->SizeOfRawData = lpSrcSection->SizeOfRawData;
	pHermit->VirtualAddress = 
		RoundUp(lpLastDestSection->VirtualAddress + lpLastDestSection->Misc.VirtualSize, lpHeader->OptionalHeader.SectionAlignment);

	pHermit->Misc.VirtualSize = lpSrcSection->Misc.VirtualSize;

	PBYTE data = (PBYTE)((DWORD) lpDestAddr + pHermit->VirtualAddress);

	memcpy(data, MAKE_PTR(lpSrcAddr, lpSrcSection->VirtualAddress), lpSrcSection->SizeOfRawData);

	return pHermit;
}

void Pack(PIMAGE_NT_HEADERS lpHeader, LPVOID lpBaseAddr, PIMAGE_SECTION_HEADER sections)
{
	srand((unsigned long)time(NULL));

	key[0] = rand();
	key[1] = rand();
	key[2] = rand();
	key[3] = rand();

	for(WORD i = 0; i < lpHeader->FileHeader.NumberOfSections; i++)
	{
		PBYTE tmpSect = new BYTE[sections[i].SizeOfRawData];
		LPVOID sectAddr = MAKE_PTR(lpBaseAddr, sections[i].VirtualAddress);

		if ((sections[i].Characteristics & IMAGE_SCN_MEM_SHARED) == 0 &&
			strcmp((const char*)sections[i].Name, ".rsrc") != 0)
		{	// non shared section and non shared resource
		
			init_sbox();
			init_sbox_key((PBYTE)&key, sizeof(key));
		
			memcpy(tmpSect, sectAddr, sections[i].SizeOfRawData);

			ZeroMemory(MAKE_PTR(lpBaseAddr, sections[i].VirtualAddress), sections[i].SizeOfRawData);

			PBYTE workmem = new BYTE[aP_workmem_size(sections[i].SizeOfRawData)];

			DWORD newSize = aP_pack(tmpSect, sectAddr, sections[i].SizeOfRawData, workmem, NULL, NULL);

			cypher_msg((LPBYTE)sectAddr, newSize);

			DWORD oldSize = sections[i].SizeOfRawData;

			sections[i].SizeOfRawData = newSize;

			//lpHeader->OptionalHeader.SizeOfImage -= oldSize;
			//lpHeader->OptionalHeader.SizeOfImage += newSize;

			if (i > 0 && sections[i].PointerToRawData != 0)
			{
				DWORD previous = 0;
				DWORD a = i-1;

				while(previous == 0 && a >= 0)
				{
					if (sections[a].PointerToRawData != 0)
						previous = RoundUp(sections[a].PointerToRawData + sections[a].SizeOfRawData, lpHeader->OptionalHeader.FileAlignment);
					else
						a--;
				}

				// align the section on disk with previous size
				sections[i].PointerToRawData =
					previous;
			}

			delete [] tmpSect;
			delete [] workmem;
		}
		else
		{	// update pointer...
			if (i > 0 && sections[i].PointerToRawData != 0)
			{
				DWORD previous = 0;
				DWORD a = i-1;

				while(previous == 0 && a >= 0)
				{
					if (sections[a].PointerToRawData != 0)
						previous = RoundUp(sections[a].PointerToRawData + sections[a].SizeOfRawData, lpHeader->OptionalHeader.FileAlignment);
					else
						a--;
				}

				// align the section on disk with previous size
				sections[i].PointerToRawData =
					previous;
			}

		}
	}

}

extern unsigned char HermitDll[7680];
//extern unsigned char HermitDll[7168];

LPVOID LoadHermit()
{
	TCHAR tmpDir[128];
	TCHAR tmpFileName[256];

	GetTempPath(128, tmpDir);

	//TEXT tmp = "tmp";

	if (GetTempFileName(tmpDir, TEXT("tmp"), 0, tmpFileName) == 0)
		return NULL;

	HANDLE hFile = CreateFile(tmpFileName, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, 0, NULL);

	if (hFile == INVALID_HANDLE_VALUE)
		return NULL;

	DWORD bytesWritten = 0;
	WriteFile(hFile, HermitDll, sizeof(HermitDll), &bytesWritten, NULL);

	CloseHandle(hFile);

	LPVOID result = InternalLoadLibrary(tmpFileName, 0);

	DeleteFile(tmpFileName);

	return result;
}

void _hermit(TCHAR* input, TCHAR* output)
{
	LPVOID lpHermitAddr = LoadHermit();

	if (lpHermitAddr == NULL)
		return;

	PIMAGE_SECTION_HEADER hermitSections = SectionHeader(lpHermitAddr);
	PIMAGE_NT_HEADERS hermitHeader = GetNTHeader(lpHermitAddr);

	LPVOID lpBaseAddr = InternalLoadLibrary(input, 10);

	PIMAGE_NT_HEADERS lpHeader = GetNTHeader(lpBaseAddr);
	WORD noSections = lpHeader->FileHeader.NumberOfSections;

	if (lpBaseAddr == NULL || lpHeader == NULL)
	{
		return;
	}

	//ExpandOptionalHeader(lpBaseAddr, 16);
	PIMAGE_DATA_DIRECTORY directory = lpHeader->OptionalHeader.DataDirectory;
	DWORD relocVirtualAddress = directory[5].VirtualAddress;
	DWORD relocSize = directory[5].Size;

	DWORD newRandBase = rand() % 0x6FFF;

	newRandBase <<= 0x10;

	ModifyReloc(
		lpBaseAddr,
		lpHeader->OptionalHeader.ImageBase,
		newRandBase,
		MAKE_PTR(lpBaseAddr, relocVirtualAddress),
		relocSize);

	lpHeader->OptionalHeader.ImageBase = newRandBase;

	Pack(lpHeader, lpBaseAddr, SectionHeader(lpBaseAddr));

	WORD sizeMirror = hermitHeader->FileHeader.NumberOfSections;
	PSECTION_MIRROR mirror = new SECTION_MIRROR[sizeMirror];
	
	for(WORD i = 0; i < hermitHeader->FileHeader.NumberOfSections; i++)
	{
		PIMAGE_SECTION_HEADER sect = _clone_section(lpBaseAddr,
			LastHeader(lpBaseAddr),
			lpHeader,
			i,
			&hermitSections[i],
			lpHermitAddr);

		lpHeader->OptionalHeader.SizeOfImage += 
			RoundUp(sect->Misc.VirtualSize, lpHeader->OptionalHeader.SectionAlignment);

		memcpy(&mirror[i].BaseHeader, &hermitSections[i], sizeof(IMAGE_SECTION_HEADER));
		memcpy(&mirror[i].NewHeader, sect, sizeof(IMAGE_SECTION_HEADER));
	}

	PIMAGE_DATA_DIRECTORY original = hermitHeader->OptionalHeader.DataDirectory;

	// export table
	DWORD eatVirtualAddress = directory[0].VirtualAddress;
	DWORD eatSize = directory[0].Size;

	directory[0].VirtualAddress = 0;
	directory[0].Size = 0;

	// IAT
	DWORD IAT_VirtualAddress = directory[1].VirtualAddress;
	DWORD IAT_Size = directory[1].Size;

	directory[1].VirtualAddress = 
		ConvertAddr(mirror, sizeMirror, original[1].VirtualAddress);

	directory[1].Size = original[1].Size;

	AdjustIAT(lpBaseAddr, mirror, sizeMirror, MAKE_PTR(lpBaseAddr, directory[1].VirtualAddress), directory[1].Size);
	
	// Resource
	DWORD rsrcVirtualAddress = directory[2].VirtualAddress;
	DWORD rsrcSize = directory[2].Size;

	//directory[2].VirtualAddress = 0;
	//directory[2].Size = 0;

	// Exception

	// Certificate

	// Relocation
	directory[5].VirtualAddress = ConvertAddr(mirror, sizeMirror, original[5].VirtualAddress);
	directory[5].Size = original[5].Size;


	AdjustRELOC(
		lpBaseAddr,
		hermitHeader->OptionalHeader.ImageBase,
		lpHeader->OptionalHeader.ImageBase,
		mirror, sizeMirror, MAKE_PTR(lpBaseAddr, directory[5].VirtualAddress), directory[5].Size);

	// Debug

	// Architecture

	// Machine

	// tls

	// load
	directory[10].VirtualAddress = 0;
	directory[10].Size = 0;

	// bound

	// iat
	DWORD iatVirtualAddress = directory[12].VirtualAddress;
	DWORD iatSize = directory[12].Size;

	directory[12].VirtualAddress = ConvertAddr(mirror, sizeMirror, original[12].VirtualAddress);
	directory[12].Size = original[12].Size;

	// delay

	// com

	// reserved

	// fix header
	DWORD AddressOfEntryPoint = lpHeader->OptionalHeader.AddressOfEntryPoint;

	lpHeader->OptionalHeader.AddressOfEntryPoint =
		ConvertAddr(mirror, sizeMirror, hermitHeader->OptionalHeader.AddressOfEntryPoint);

	DWORD *ptrToWriteKey = (DWORD*)MAKE_PTR(lpBaseAddr, lpHeader->OptionalHeader.AddressOfEntryPoint);
	
	ptrToWriteKey[2] = key[0];
	ptrToWriteKey[3] = key[1];
	ptrToWriteKey[4] = key[2];
	ptrToWriteKey[5] = key[3];
	ptrToWriteKey[6] = eatVirtualAddress;
	ptrToWriteKey[7] = eatSize;
	ptrToWriteKey[8] = IAT_VirtualAddress;
	ptrToWriteKey[9] = IAT_Size;
	ptrToWriteKey[10] = relocVirtualAddress;
	ptrToWriteKey[11] = relocSize;
	ptrToWriteKey[12] = rsrcVirtualAddress;
	ptrToWriteKey[13] = rsrcSize;
	ptrToWriteKey[14] = AddressOfEntryPoint;
	WORD *ptrToWriteKeyWord = (WORD *) &ptrToWriteKey[15];
	*ptrToWriteKeyWord = noSections;	// Write in hermit old no. sections

	InternalWriteLibraryToFile(lpBaseAddr, output);

	return;
}

/**
 *
 **/
void _polymer(TCHAR *argv1)
{
	_hermit(argv1, argv1);
}

#ifdef __EXE__

void _polymer(TCHAR *argv1, TCHAR *argv2)
{
	_hermit(argv1, argv2);
}
#endif

